import { assertNotFull, creatorsToJsonMetadataCreators, assertCanAdd } from './Client.helpers.mjs';
import { CandyMachineToUpdateNotFoundError } from '../../errors/CandyMachineError.mjs';
import { randomStr } from '../../utils/common.mjs';

// uploadAssetForCandyMachine (single)
// -----------------

async function uploadAssetForCandyMachine(params) {
  var _params$properties$cr, _params$properties, _params$seller_fee_ba;

  const candyMachine = await this.findByAddress(params.candyMachineAddress);

  if (candyMachine == null) {
    throw new CandyMachineToUpdateNotFoundError(params.candyMachineAddress);
  }

  assertNotFull(candyMachine, candyMachine.assetsCount); // Default NFT creators to equal those of the Candy Machine

  const creators = (_params$properties$cr = (_params$properties = params.properties) === null || _params$properties === void 0 ? void 0 : _params$properties.creators) !== null && _params$properties$cr !== void 0 ? _params$properties$cr : candyMachine.creators;
  const uploadProperties = { ...params.properties,
    creators: creatorsToJsonMetadataCreators(creators)
  }; // TODO(thlorenz): Is this correct?

  const seller_fee_basis_points = (_params$seller_fee_ba = params.seller_fee_basis_points) !== null && _params$seller_fee_ba !== void 0 ? _params$seller_fee_ba : candyMachine.sellerFeeBasisPoints;
  const {
    candyMachineAddress,
    authoritySigner,
    properties,
    // TODO(thlorenz): prevent same asset from being uploaded twice, remove once
    // API improves to have clearly separated properties
    assets,
    ...uploadInputParams
  } = params;
  const uploadInput = { ...uploadInputParams,
    properties: uploadProperties,
    seller_fee_basis_points
  };
  const {
    uri,
    metadata
  } = await this.metaplex.nfts().uploadMetadata(uploadInput);
  let addAssetsTransactionId;

  if (params.addToCandyMachine) {
    const assetName = selectAssetName(metadata, params);
    const {
      transactionId
    } = await this.addAssets({
      candyMachineAddress: params.candyMachineAddress,
      authoritySigner: params.authoritySigner,
      assets: [{
        uri,
        name: assetName
      }],
      confirmOptions: params.confirmOptions
    });
    addAssetsTransactionId = transactionId;
  }

  return {
    metadata,
    uri,
    addAssetsTransactionId
  };
} // -----------------
// uploadAssetsForCandyMachine (multiple)
// -----------------
// NOTE: the `[key: string]: unknown;` that is part of `UploadMetadataInput` breaks type safety once
// we extend a derived type, thus we don't get intellisense for `CandyMachineAsset` either

async function uploadAssetsForCandyMachine(params) {
  const candyMachine = await this.findByAddress(params.candyMachineAddress);

  if (candyMachine == null) {
    throw new CandyMachineToUpdateNotFoundError(params.candyMachineAddress);
  }

  assertNotFull(candyMachine, candyMachine.assetsCount);
  assertCanAdd(candyMachine, candyMachine.assetsCount, params.assets.length);
  const {
    parallel = false,
    addToCandyMachine = false
  } = params; // TODO(thlorenz): prevent same asset from being uploaded twice, remove once
  // API improves to have clearly separated properties

  const {
    assets,
    ...assetLessParams
  } = params;
  const uploadParams = assets.map(x => {
    const param = { ...assetLessParams,
      image: x,
      name: x.displayName,
      // We add them all in one transaction after all assets are uploaded
      addToCandyMachine: false
    };
    return param;
  });
  let uploadedAssets = [];
  const errors = [];

  if (parallel) {
    // NOTE: we are uploading in parallel here but if only one upload was to fail
    // all the other ones still happen as we cannot cancel promises
    const promises = uploadParams.map(async assetParam => {
      let uploaded;
      let err;

      try {
        uploaded = await _uploadAssetAndSelectName(this, params, assetParam);
      } catch (e) {
        errors.push(e);
      }

      return {
        uploaded,
        err
      };
    });
    const results = await Promise.all(promises);

    for (const {
      err,
      uploaded
    } of results) {
      if (err) {
        errors.push(err);
      } else {
        uploadedAssets.push(uploaded);
      }
    }
  } else {
    for (const assetParam of uploadParams) {
      try {
        uploadedAssets.push(await _uploadAssetAndSelectName(this, params, assetParam));
      } catch (err) {
        errors.push(err);
        continue;
      }
    }
  }

  let addAssetsTransactionId;
  let updatedCandyMachine = candyMachine;

  if (addToCandyMachine && uploadedAssets.length > 0) {
    const configLines = uploadedAssets.map(x => ({
      uri: x.uri,
      name: x.name
    }));
    const {
      transactionId,
      candyMachine
    } = await this.addAssets({ ...params,
      assets: configLines
    });
    addAssetsTransactionId = transactionId;

    if (candyMachine != null) {
      updatedCandyMachine = candyMachine;
    }
  }

  return {
    addAssetsTransactionId,
    uploadedAssets,
    candyMachine: updatedCandyMachine,
    errors
  };
}

async function _uploadAssetAndSelectName(candyMachine, params, asset) {
  const {
    uri,
    metadata
  } = await candyMachine.uploadAssetForCandyMachine({ ...params,
    ...asset
  });
  return {
    uri,
    metadata,
    name: selectAssetName(metadata, asset)
  };
}

function selectAssetName(metadata, asset) {
  var _metadata$name;

  return (_metadata$name = metadata.name) !== null && _metadata$name !== void 0 ? _metadata$name : typeof asset.name === 'string' ? asset.name : randomStr();
}

export { uploadAssetForCandyMachine, uploadAssetsForCandyMachine };
//# sourceMappingURL=Client.upload.mjs.map
